/*
								+----------------------------------+
								|                                  |
								|  ***  berLame GLUT window  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#include "../berLame_src/CallStack.h"
#if defined(WIN32) || defined(_WIN64)
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/Vector.h"

static int n_width = 512, n_height = 512;
// window size

static CGLState *p_state = 0;
// OpenGL state guard

static CGLTexture_2D *p_equations = 0;
// texture with equation coefficients

static CGLShader *p_shader = 0;
static int n_texture_param_register = -1;
// shader to solve the equations

static const float p_vertex[] = {0, 0, -1, -1, 0,
	1, 0, 1, -1, 0, 1, 1, 1, 1, 0, 0, 1, -1, 1, 0};
// fullscreen quad vertex coordinates

static bool b_show_results = true;
// switches between results and equation coefficients display

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera
}

// here we render frame
void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if(b_show_results) {
		p_equations->Bind_Enable(p_state);
		p_shader->Bind(p_state); // enable shader
		p_shader->SetParam1i(p_state, n_texture_param_register, 0); // texture is in unit 0
		// use shader to show equations sollutions
	} else {
		p_equations->Bind_Enable(p_state);
		p_state->BindProgramObject(0); // disable shader
		// show equations texture
	}

	glInterleavedArrays(GL_T2F_V3F, 0, p_vertex);
	glDrawArrays(GL_QUADS, 0, 4);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	// draw fullscreen quad

	glutSwapBuffers();
	// display results
}

void Cleanup()
{
	if(p_shader)
		delete p_shader;
	if(p_equations)
		delete p_equations;
	if(p_state)
		delete p_state;
}

void onKeyPressed(unsigned char n_key_code, int n_x, int n_y)
{
	if(n_key_code == 27) {
		Cleanup();
		exit(0);
	} else if(n_key_code == ' ') {
		b_show_results = !b_show_results;
		glutPostRedisplay();
	}
}

float *p_GenerateEquationCoefficients(int n_size_x, int n_size_y)
{
	float *p_coeff;
	if(!(p_coeff = new float[3 * n_size_x * n_size_y]))
		return 0;

	float *p_coeff_ptr = p_coeff;
	for(int y = 0; y < n_size_y; ++ y) {
		for(int x = 0; x < n_size_x; ++ x, p_coeff_ptr += 3) {
			Vector3f v_dir(float(x - n_size_x / 2) * 2 / n_size_x,
				float(y - n_size_y / 2) / n_size_y * 2, 1);
			Vector3f v_org(0, 0, -.5f);
			// ray fired from camera

			Vector3f v_center(0, 0, 0);
			float f_radius = .33f;
			// sphere data

			float a = v_dir.f_Dot(v_dir);
			Vector3f v_conn = v_org - v_center;
			float b = 2 * v_dir.f_Dot(v_conn);
			float c = v_conn.f_Dot(v_conn) - f_radius * f_radius;
			// quadratic equation coefficients

			p_coeff_ptr[0] = a;
			p_coeff_ptr[1] = b;
			p_coeff_ptr[2] = c;
		}
	}

	return p_coeff;
}

int main(int n_arg_num, const char **p_arg_list)
{
	glutInit(&n_arg_num, (char**)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutInitWindowPosition(200, 200);
	glutCreateWindow("berLame GLUT window");
	// init OpenGL using GLUT

	if(!CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_float")) {
		fprintf(stderr, "error: ARB_texture_float is not supported\n");
		Cleanup();
		return -1;
	}
	// need ARB_texture_float extension to create floating-point texture

	if(!(p_state = new CGLState())) {
		fprintf(stderr, "error: not enough memory\n");
		Cleanup();
		return -1;
	}
	// create OpenGL state guard

	float *p_coeffs;
	if(!(p_coeffs = p_GenerateEquationCoefficients(n_width, n_height))) {
		fprintf(stderr, "error: not enough memory\n");
		Cleanup();
		return -1;
	}
	// generate equation coefficients

	if(!(p_equations = new CGLTexture_2D(p_state, n_width, n_height, GL_RGB16F_ARB,
	   false, 0, GL_RGB, GL_FLOAT, p_coeffs)) || !p_equations->b_Status()) {
		fprintf(stderr, "error: failed to create equations texture\n");
		delete[] p_coeffs;
		Cleanup();
		return -1;
	}
	// create floating-point RGB texture, containing equation coefficients

	delete[] p_coeffs;
	// don't need this anymore

    const char *p_s_solve_quad =
		"uniform sampler2D n_texture;\n"
		"\n"
		"void main()\n"
		"{\n"
		"    vec3 texel = texture2D(n_texture, gl_TexCoord[0].st).rgb; // read texture\n"
		"\n"
		"    float a = texel.r;\n"
		"    float b = texel.g;\n"
		"    float c = texel.b; // get quadratic equation coefficients\n"
		"\n"
		"    float D = b * b - 4.0 * a * c; // calculate discriminant squared\n"
		"\n"
		"    if(D < 0.0) {\n"
		"        gl_FragColor = vec4(0); // equation has no sollution\n"
		"    } else {\n"
		"        D = sqrt(D); // calculate discriminant\n"
		"\n"
		"        float x1 = (-b - D) / (2.0 * a);\n"
		"        float x2 = (-b + D) / (2.0 * a); // calculate roots\n"
		"\n"
		"        gl_FragColor = vec3(x1, x2, 1.0).xyzz; // equation has one or two sollutions\n"
		"    }\n"
		"}\n";

    if(!(p_shader = new CGL_ARB_shader(0, p_s_solve_quad))) {
        fprintf(stderr, "error: not enough memory\n");
		Cleanup();
        return -1;
    }
	// create shader for solving quadratic equations

    char *p_s_info_log = 0;
    if(!p_shader->Compile(p_state, p_s_info_log)) {
        fprintf(stderr, "error: error compiling shader:\n%s\n",
            (p_s_info_log)? p_s_info_log : "(null)");
        if(p_s_info_log)
            delete[] p_s_info_log;
        return -1;
    }
    if(p_s_info_log) {
        fprintf(stderr, "warning: while compiling shader:\n%s\n", p_s_info_log);
        delete[] p_s_info_log;
        p_s_info_log = 0;
    }
    if(!p_shader->Link(p_s_info_log)) {
        fprintf(stderr, "error: error linking shader:\n%s\n",
            (p_s_info_log)? p_s_info_log : "(null)");
        if(p_s_info_log)
            delete[] p_s_info_log;
        return -1;
    }
    if(p_s_info_log) {
        fprintf(stderr, "warning: while linking shader:\n%s\n", p_s_info_log);
        delete[] p_s_info_log;
        p_s_info_log = 0;
    }
    // compile and link the shader

	if((n_texture_param_register = p_shader->n_GetUniformLocation("n_texture")) < 0) {
        fprintf(stderr, "error: can't find \'n_texture\' parameter\n");
		Cleanup();
        return -1;
	}
	// find index of register on GPU where "n_texture" uniform is stored

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutKeyboardFunc(onKeyPressed);
	// register our callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
